"""
Reads data from FPT (memo) files.

FPT files are used to varying lenght text or binary data which is too
large to fit in a DBF field.

VFP == Visual FoxPro
DB3 == dBase III
DB4 == dBase IV
"""
from collections import namedtuple
from .ifiles import ifind
from .struct_parser import StructParser



VFPFileHeader = StructParser(
    'FPTHeader',
    '>LHH504s',
    ['nextblock',
     'reserved1',
     'blocksize',
     'reserved2'])

VFPMemoHeader = StructParser(
    'FoxProMemoHeader',
    '>LL',
    ['type',
     'length'])

DB4MemoHeader = StructParser(
    'DBase4MemoHeader',
    '<LL',
    ['reserved',  # Always 0xff 0xff 0x08 0x08.
     'length'])

# Used for Visual FoxPro memos to distinguish binary from text memos.

class VFPMemo(bytes):
    pass

class BinaryMemo(VFPMemo):
    pass

class PictureMemo(BinaryMemo):
    pass

class ObjectMemo(BinaryMemo):
    pass

class TextMemo(VFPMemo):
    pass


VFP_TYPE_MAP = {
    0x0: PictureMemo,
    0x1: TextMemo,
    0x2: ObjectMemo,
}


class MemoFile(object):
    def __init__(self, filename):
        self.filename = filename
        self._open()
        self._init()

    def _init(self):
        pass

    def _open(self):
        self.file = open(self.filename, 'rb')
        # Shortcuts for speed.
        self._read = self.file.read
        self._seek = self.file.seek

    def _close(self):
        self.file.close()

    def __getitem__(self, index):
        raise NotImplemented

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._close()
        return False


class FakeMemoFile(MemoFile):
    def __getitem__(self, i):
        return None

    def _open(self):
        pass

    _init = _close = _open


class VFPMemoFile(MemoFile):
    def _init(self):
        self.header = VFPFileHeader.read(self.file)

    def __getitem__(self, index):
        """Get a memo from the file."""
        if index <= 0:
            return None

        self._seek(index * self.header.blocksize)
        memo_header = VFPMemoHeader.read(self.file)

        data = self._read(memo_header.length)
        if len(data) != memo_header.length:
            raise IOError('EOF reached while reading memo')
        
        return VFP_TYPE_MAP.get(memo_header.type, BinaryMemo)(data)


class DB3MemoFile(MemoFile):
    """dBase III memo file."""
    # Code from dbf.py
    def __getitem__(self, index):
        if index <= 0:
            return None

        block_size = 512
        self._seek(index * block_size)
        data = b''
        while True:
            newdata = self._read(block_size)
            if not newdata:
                return data
            data += newdata

            # Todo: some files (help.dbt) has only one field separator.
            # Is this enough for all file though?
            end_of_memo = data.find(b'\x1a')
            if end_of_memo != -1:
                return data[:end_of_memo]

            # Alternative end of memo markers:
            # \x1a\x1a
            # \x0d\x0a

        return data[:eom]        

class DB4MemoFile(MemoFile):
    """dBase IV memo file"""
    def __getitem__(self, index):
        if index <= 0:
            return None

        # Todo: read this from the file header.
        block_size = 512

        self._seek(index * block_size)
        memo_header = DB4MemoHeader.read(self.file)
        data = self._read(memo_header.length)
        # Todo: fields are terminated in different ways.
        # \x1a is one of them
        # \x1f seems to be another (dbase_8b.dbt)
        return data.split(b'\x1f', 1)[0]


def find_memofile(dbf_filename):
    for ext in ['.fpt', '.dbt']:
        name = ifind(dbf_filename, ext=ext)
        if name:
            return name
    else:
        return None


def open_memofile(filename, dbversion):
    if filename.lower().endswith('.fpt'):
        return VFPMemoFile(filename)
    else:
        # print('######', dbversion)
        if dbversion == 0x83:
            return DB3MemoFile(filename)
        else:
            return DB4MemoFile(filename)



"""
Read DBF files with Python.

Example:

    >>> from dbfread import DBF
    >>> for record in DBF('people.dbf'):
    ...     print(record)
    OrderedDict([('NAME', 'Alice'), ('BIRTHDATE', datetime.date(1987, 3, 1))])
    OrderedDict([('NAME', 'Bob'), ('BIRTHDATE', datetime.date(1980, 11, 12))])

Full documentation at https://dbfread.readthedocs.io/

"""
__author__ = 'Ole Martin Bjorndalen'
__email__ = 'ombdalen@gmail.com'
__url__ = 'https://dbfread.readthedocs.io/'
__license__ = 'MIT'

from .dbf import DBF
from .deprecated_dbf import open, read
from .exceptions import *
from .field_parser import FieldParser, InvalidValue
from .version import version_info, version as __version__

# Prevent splat import.
__all__ = []



from __future__ import print_function
import sys
import warnings
from .dbf import DBF

class DeprecatedDBF(DBF, list):
    """This is the old version of the table which is a subclass of list.

    It is included for backwards compatability with 1.0 and older."""
    @property
    def loaded(self):
        # Since records are loaded into the table object
        # we have to check the deleted attribute here.
        return isinstance(self._deleted, list)

    def load(self):
        if not self.loaded:
            self[:] = self._iter_records(b' ')
            self._deleted = list(self._iter_records(b'*'))

    def unload(self):
        # self.loaded is not checked here because this
        # is called by __init__() where self.loaded=False.
        # Also, unloading twice has no consequences.
        del self[:]
        self._deleted = None
        
    def __iter__(self):
        if self.loaded:
            return list.__iter__(self)
        else:
            return self._iter_records()

    def __len__(self):
        if self.loaded:
            return list.__len__(self)
        else:
            return self._count_records()

    def __repr__(self):
        if self.loaded:
            return list.__repr__(self)
        else:
            return '<unloaded DBF table {!r}>'.format(self.filename)

def read(filename, load=True, **kwargs):
    warnings.warn("dbfread.read() has been replaced by DBF(load=True)"
                  " and will be removed in 2.2.")
    return DeprecatedDBF(filename, load=True, **kwargs)

def open(filename, load=True, **kwargs):
    warnings.warn("dbfread.open() has been replaced by DBF()"
                  " and will be removed in 2.2.")
    return DeprecatedDBF(filename, load=True, **kwargs)



"""
Tests reading from database.
"""
from pytest import fixture
import datetime
from .dbf import DBF

@fixture
def table():
    return DBF('testcases/memotest.dbf')

@fixture
def loaded_table():
    return DBF('testcases/memotest.dbf', load=True)

# This relies on people.dbf having this exact content.
records = [{u'NAME': u'Alice',
            u'BIRTHDATE': datetime.date(1987, 3, 1),
            u'MEMO': u'Alice memo'},
           {u'NAME': u'Bob',
            u'BIRTHDATE': datetime.date(1980, 11, 12),
            u'MEMO': u'Bob memo'}]
deleted_records = [{u'NAME': u'Deleted Guy',
                    u'BIRTHDATE': datetime.date(1979, 12, 22),
                    u'MEMO': u'Deleted Guy memo'}]

def test_len():
    assert len(table()) == 2
    assert len(table().deleted) == 1

    assert len(loaded_table()) == 2
    assert len(loaded_table().deleted) == 1

def test_list():
    assert list(table()) == records
    assert list(table().deleted) == deleted_records
    
    assert list(loaded_table()) == records
    assert list(loaded_table().deleted) == deleted_records

    # This should not return old style table which was a subclass of list.
    assert not isinstance(table(), list)



# Table from dbf.py by Ethan Furman
codepages = {
    0x00: ('ascii', "plain ol' ascii"),
    0x01: ('cp437', 'U.S. MS-DOS'),
    0x02: ('cp850', 'International MS-DOS'),
    0x03: ('cp1252', 'Windows ANSI'),
    0x04: ('mac_roman', 'Standard Macintosh'),
    0x08: ('cp865', 'Danish OEM'),
    0x09: ('cp437', 'Dutch OEM'),
    0x0A: ('cp850', 'Dutch OEM (secondary)'),
    0x0B: ('cp437', 'Finnish OEM'),
    0x0D: ('cp437', 'French OEM'),
    0x0E: ('cp850', 'French OEM (secondary)'),
    0x0F: ('cp437', 'German OEM'),
    0x10: ('cp850', 'German OEM (secondary)'),
    0x11: ('cp437', 'Italian OEM'),
    0x12: ('cp850', 'Italian OEM (secondary)'),
    0x13: ('cp932', 'Japanese Shift-JIS'),
    0x14: ('cp850', 'Spanish OEM (secondary)'),
    0x15: ('cp437', 'Swedish OEM'),
    0x16: ('cp850', 'Swedish OEM (secondary)'),
    0x17: ('cp865', 'Norwegian OEM'),
    0x18: ('cp437', 'Spanish OEM'),
    0x19: ('cp437', 'English OEM (Britain)'),
    0x1A: ('cp850', 'English OEM (Britain) (secondary)'),
    0x1B: ('cp437', 'English OEM (U.S.)'),
    0x1C: ('cp863', 'French OEM (Canada)'),
    0x1D: ('cp850', 'French OEM (secondary)'),
    0x1F: ('cp852', 'Czech OEM'),
    0x22: ('cp852', 'Hungarian OEM'),
    0x23: ('cp852', 'Polish OEM'),
    0x24: ('cp860', 'Portuguese OEM'),
    0x25: ('cp850', 'Portuguese OEM (secondary)'),
    0x26: ('cp866', 'Russian OEM'),
    0x37: ('cp850', 'English OEM (U.S.) (secondary)'),
    0x40: ('cp852', 'Romanian OEM'),
    0x4D: ('cp936', 'Chinese GBK (PRC)'),
    0x4E: ('cp949', 'Korean (ANSI/OEM)'),
    0x4F: ('cp950', 'Chinese Big 5 (Taiwan)'),
    0x50: ('cp874', 'Thai (ANSI/OEM)'),
    0x57: ('cp1252', 'ANSI'),
    0x58: ('cp1252', 'Western European ANSI'),
    0x59: ('cp1252', 'Spanish ANSI'),
    0x64: ('cp852', 'Eastern European MS-DOS'),
    0x65: ('cp866', 'Russian MS-DOS'),
    0x66: ('cp865', 'Nordic MS-DOS'),
    0x67: ('cp861', 'Icelandic MS-DOS'),
    # 0x68: (None, 'Kamenicky (Czech) MS-DOS'),
    # 0x69: (None, 'Mazovia (Polish) MS-DOS'),
    0x6a: ('cp737', 'Greek MS-DOS (437G)'),
    0x6b: ('cp857', 'Turkish MS-DOS'),
    0x78: ('cp950', 'Traditional Chinese '
                    '(Hong Kong SAR, Taiwan) Windows'),
    0x79: ('cp949', 'Korean Windows'),
    0x7a: ('cp936', 'Chinese Simplified (PRC, Singapore) Windows'),
    0x7b: ('cp932', 'Japanese Windows'),
    0x7c: ('cp874', 'Thai Windows'),
    0x7d: ('cp1255', 'Hebrew Windows'),
    0x7e: ('cp1256', 'Arabic Windows'),
    0xc8: ('cp1250', 'Eastern European Windows'),
    0xc9: ('cp1251', 'Russian Windows'),
    0xca: ('cp1254', 'Turkish Windows'),
    0xcb: ('cp1253', 'Greek Windows'),
    0x96: ('mac_cyrillic', 'Russian Macintosh'),
    0x97: ('mac_latin2', 'Macintosh EE'),
    0x98: ('mac_greek', 'Greek Macintosh'),
}


def guess_encoding(language_driver):
    if language_driver in codepages:
        return codepages[language_driver][0]
    else:
        raise LookupError('Unable to guess encoding '
                          'for languager driver byte '
                          '0x{:x}'.format(language_driver))



"""
Functions for dealing with mixed-case files from case-preserving file
systems.

Todo:

  - handle patterns that already have brackets
"""

from __future__ import print_function
import os
import glob
import fnmatch


def ipat(pat):
    """Convert glob pattern to case insensitive form."""

    (dirname, pat) = os.path.split(pat)

    # Convert '/path/to/test.fpt' => '/path/to/[Tt][Ee][Ss][Tt].[]' 
    newpat = ''
    for c in pat:
        if c.isalpha:
            u = c.upper()
            l = c.lower()
            if u != l:
                newpat = newpat + '[' + u + l + ']'
            else:
                newpat += c
        else:
            newpat += c

    newpat = os.path.join(dirname, newpat)

    return newpat

def ifnmatch(name, pat):
    """Case insensitive version of fnmatch.fnmatch()"""
    return fnmatch.fnmatch(name, ipat(pat))

def iglob(pat):
    """Case insensitive version of glob.glob()"""
    return glob.glob(ipat(pat))

def ifind(pat, ext=None):
    """Look for a file in a case insensitive way.

    Returns filename it a matching file was found, or None if it was not.
    """

    if ext:
        pat = os.path.splitext(pat)[0] + ext

    files = iglob(pat)
    if files:
        return files[0]  # Return an arbitrary file
    else:
        return None

__all__ = ['ipat', 'ifnmatch', 'iglob', 'ifind']
